# final_code.py

"Starting final_code script"

# Raspberry Pi Pico - Scarecase code

import board
import math, time
import pwmio
import digitalio
from digitalio import DigitalInOut, Direction, Pull

#--------------------------------------------------------------------------------
# Class to represent a single hardware hobby servo.  This wraps up all the
# configuration information and current state into a single object.  The
# creation of the object also initializes the physical pin hardware.

# This assumes a tiny 9G servo has been wired up to the Pico as follows:
#   Pico pin 40 (VBUS)  -> servo red   (+5V)
#   Pico pin 38 (GND)   -> servo brown (GND)
#   Pico pin 1  (GP0)   -> servo orange (SIG)

class Servo():
    def __init__(self, pin, pulse_rate=50):
        # Create a PWMOut object on the desired pin to drive the servo.
        # Note that the initial duty cycle of zero generates no pulses, which
        # for many servos will present as a quiescent low-power state.
        self.pwm = pwmio.PWMOut(board.GP0, duty_cycle=0, frequency=pulse_rate)

        # Save the initialization arguments within the object for later reference.
        self.pin = pin
        self.pulse_rate = pulse_rate

        # Initialize the other state variables.
        self.target = None    # target angle; None indicates the "OFF" state
        self.debug = False    # flag to control print output

    def write(self, angle):
        # Object method to issue a servo command by updating the PWM signal
        # output.  This function maps an angle specified in degrees between 0
        # and 180 to a servo command pulse width between 1 and 2 milliseconds,
        # and then to the corresponding duty cycle fraction, specified as a
        # 16-bit fixed-point integer.  As a special case, an angle value of None
        # will turn off the output; many servos will then become backdrivable.

        # calculate the desired pulse width in units of seconds
        if angle is None:
            pulse_width = 0.0
        else:
            pulse_width  = 0.001 + angle * (0.001 / 180.0)

        # calculate the duration in seconds of a single pulse cycle
        cycle_period = 1.0 / self.pulse_rate

        # calculate the desired ratio of pulse ON time to cycle duration
        duty_cycle   = pulse_width / cycle_period

        # convert the ratio into a 16-bit fixed point integer
        duty_fixed   = int(2**16 * duty_cycle)

        # limit the ratio range and apply to the hardware driver
        self.pwm.duty_cycle = min(max(duty_fixed, 0), 65535)

        # save the target value in the object attribute (i.e. variable)
        self.target = angle

        # if requested, print some diagnostics to the console
        if self.debug:
            print(f"Driving servo to angle {angle}")
            print(f" Pulse width {pulse_width} seconds")
            print(f" Duty cycle {duty_cycle}")
            print(f" Command value {self.pwm.duty_cycle}\n")

#--------------------------------------------------------------------------------
# Movement primitive to smoothly move from a start to end angle at a constant rate.
# It does not return until the movement is complete.
# The angles are specified in degrees, the speed in degrees/second.

def linear_move(servo, start, end, speed=60, update_rate=50):
    # Calculate the number of seconds to wait between target updates to allow
    # the motor to move.
    # Units:  seconds = 1.0 / (cycles/second)
    interval = 1.0 / update_rate

    # Compute the size of each step in degrees.
    # Units:  degrees = (degrees/second) * second
    step = speed * interval

    # Output the start angle once before beginning the loop.  This guarantees at
    # least one angle will be output even if the start and end are equal.
    angle = start
    servo.write(angle)

    # Loop once for each incremental angle change.
    while angle != end:
        time.sleep(interval)            # pause for the sampling interval

        # Update the target angle.  The positive and negative movement directions
        # are treated separately.
        if end >= start:
            angle += step;              # movement in the positive direction
            if angle > end:
                angle = end             # end at an exact position
        else:
            angle -= step               # movement in the negative direction
            if angle < end:
                angle = end             # end at an exact position

        servo.write(angle)              # update the hardware

#--------------------------------------------------------------------------------
# This module provides a class for measuring distance using an HC-04 ultrasonic
# ranger.

# The device requires one digital trigger output and one digital echo input.
# It is a 5V device, so echo needs to be level-shifted to 3.3V logic levels.  One option
# is a pair of resistors as a voltage divider.

# This implementation uses bit-banged digital I/O to measure the echo pulse
# width.  But the timing precision is limited by CircuitPython execution and
# clock resolution, so the result is only good for rough proximity detection.
class Sonar:
    def __init__(self, ECHO=board.GP20, TRIG=board.GP21):
        """This class represents an HC-04 ultraonic ranger module.  It uses one output
        pin for trigger and one input pin for echo pulse measurement.  N.B. the
        device is 5V so the ECHO pin will need level-shifting to 3.3V logic
        levels."""

        self._trig = digitalio.DigitalInOut(TRIG)
        self._trig.direction = digitalio.Direction.OUTPUT
        self._trig.value = False

        self._echo = digitalio.DigitalInOut(ECHO)
        self._echo.direction = digitalio.Direction.INPUT

    def ping_sync(self):
        """Measure the round-trip echo time, returning either a value in seconds or None
        if the measurement times out.  N.B. this function will not return until
        the cycle is done, so it is not compatible with asynchronous event
        loops.
        """

        # trigger the ultrasonic pulse
        self._trig.value = True
        time.sleep(1e-5)
        self._trig.value = False

        start = time.monotonic_ns()
        timeout = start + int(0.1 * 1e9)

        # wait for echo to go high
        while self._echo.value is False:
            now = time.monotonic_ns()
            if now > timeout:
                # print("Error: ECHO never went high.")
                return None

        # wait for echo to go low
        while True:
            now = time.monotonic_ns()

            if self._echo.value is False:
                return 1e-9 * (now - start)

            if now >= timeout:
                return None

    def deinit(self):
        """Manage resource release as part of object lifecycle."""
        self._trig.deinit()
        self._echo.deinit()
        self._trig  = None
        self._echo = None

    def __enter__(self):
        return self

    def __exit__(self):
        # Automatically deinitializes the hardware when exiting a context.
        self.deinit()

#--------------------------------------------------------------------------------
# This assumes a Pololu DRV8833 dual motor driver has been wired up to the Pico as follows:
#   Pico pin 24, GPIO18   -> AIN1
#   Pico pin 25, GPIO19   -> AIN2
#   Pico pin 26, GPIO20   -> BIN2
#   Pico pin 27, GPIO21   -> BIN1
#   any Pico GND          -> GND

# Class to represent a single dual H-bridge driver.
class DRV8833():
    def __init__(self, AIN1=board.GP18, AIN2=board.GP19, BIN2=board.GP20, BIN1=board.GP21, pwm_rate=20000):
        # Create a pair of PWMOut objects for each motor channel.
        self.ain1 = pwmio.PWMOut(AIN1, duty_cycle=0, frequency=pwm_rate)
        self.ain2 = pwmio.PWMOut(AIN2, duty_cycle=0, frequency=pwm_rate)

        self.bin1 = pwmio.PWMOut(BIN1, duty_cycle=0, frequency=pwm_rate)
        self.bin2 = pwmio.PWMOut(BIN2, duty_cycle=0, frequency=pwm_rate)

    def write(self, channel, rate):
        """Set the speed and direction on a single motor channel.

        :param channel:  0 for motor A, 1 for motor B
        :param rate: modulation value between -1.0 and 1.0, full reverse to full forward."""

        # convert the rate into a 16-bit fixed point integer
        pwm = min(max(int(2**16 * abs(rate)), 0), 65535)

        if channel == 0:
            if rate < 0:
                self.ain1.duty_cycle = pwm
                self.ain2.duty_cycle = 0
            else:
                self.ain1.duty_cycle = 0
                self.ain2.duty_cycle = pwm
        else:
            if rate < 0:
                self.bin1.duty_cycle = pwm
                self.bin2.duty_cycle = 0
            else:
                self.bin1.duty_cycle = 0
                self.bin2.duty_cycle = pwm

#---------------------------------------------------------------
# create objects
# DC DRIVER
print("Creating driver object.")
driver = DRV8833()
ON_RATTLE_POWER = 1.0
MOTOR_1 = 0
MOTOR_2 = 1
MIN_RATTLE_DURATION = 30
rattle_duration = 0

def isRattleDone():
    return rattle_duration > MIN_RATTLE_DURATION

def startRattle():
    driver.write(MOTOR_1, ON_RATTLE_POWER)
    driver.write(MOTOR_2, ON_RATTLE_POWER)

def stopRattle():
    driver.write(MOTOR_1, 0.0)
    driver.write(MOTOR_2, 0.0)

# HINGE SERVO
print("Creating hinge_servo object.")
hinge_servo = Servo(board.GP0)
HINGE_CLOSED_DEGREES = 210.0 # NEED TO UPDATE
HINGE_OPEN_DEGREES = -50.0 # NEED TO UPDATE

def setHingeToClose():
    hinge_servo.write(HINGE_CLOSED_DEGREES)

def fastOpenHinge():
    hinge_servo.write(HINGE_OPEN_DEGREES)

def slowCloseHinge():
    linear_move(hinge_servo, HINGE_OPEN_DEGREES, HINGE_CLOSED_DEGREES, speed=45)

# LID SONAR
#SMALL ROUNDTRIP-----CLOSED (0.000976) ------OPENED (0.00615) ---FULLY OPENED (0.00195) ---LARGE ROUNDTRIP#
lid_sonar = Sonar(ECHO=board.GP14, TRIG=board.GP15)
LID_CLOSED_ROUNDTRIP = 0.000976
LID_DELTA = 0.0005
last_couple_lids = []
lid_time = 0.0

def isLidClosed():
    return lid_time <= LID_CLOSED_ROUNDTRIP + LID_DELTA

def isLidOpened():
    return lid_time >= LID_CLOSED_ROUNDTRIP + LID_DELTA

# PERSON SONAR
#SMALL ROUNDTRIP---CLOSE (0.00488281) --- FAR (0.00585938) ---LARGE ROUNDTRIP#
person_sonar = Sonar(ECHO=board.GP16, TRIG=board.GP17)
FAR_PERSON_ROUNDTRIP = 0.0058
PERSON_DELTA = 0.0004
last_couple_persons = []
person_time = 0.0

def isPersonClose():
    return person_time <= FAR_PERSON_ROUNDTRIP - PERSON_DELTA

def isPersonFar():
    return person_time >= FAR_PERSON_ROUNDTRIP - PERSON_DELTA

# SONAR
def updateModeTime(sonar, last_couple):
    curr_time = sonar.ping_sync()
    last_couple.append(curr_time)
    if len(last_couple) > 10:
        last_couple.pop(0)
    return max(set(last_couple), key=last_couple.count)

# PIR MOTION
pir_motion = DigitalInOut(board.GP11)
pir_motion.direction = Direction.INPUT
is_motion_detected = False

def isMotionDetected():
    return pir_motion.value is True

#---------------------------------------------------------------
# other constants
MAX_WAIT_FOR_OPEN_DURATION = 50
wait_for_open_duration = 0

def isWaitForOpenDone():
    return wait_for_open_duration > MAX_WAIT_FOR_OPEN_DURATION

SCARE_DURATION = 5.0
SLEEP_AFTER_SCARE = 10.0

#---------------------------------------------------------------
# initialize
state_index = "idle"

stopRattle()
setHingeToClose()

# measure closed lid time
for i in range(30):
    lid_time = updateModeTime(lid_sonar, last_couple_lids)
    print(lid_time)

LID_CLOSED_ROUNDTRIP = lid_time
print("LID_CLOSED_ROUNDTRIP = " + str(LID_CLOSED_ROUNDTRIP))

while (True):
    lid_time = updateModeTime(lid_sonar, last_couple_lids)
    person_time = updateModeTime(person_sonar, last_couple_persons)
    is_motion_detected = isMotionDetected()

    print(state_index + ", " + str(is_motion_detected) + ", " + str(person_time) + ", " + str(lid_time))

    if state_index == "idle":
        # if motion detected, begin rattling to attract
        if pir_motion.value is True:
            print("MOTION DETECTED!")
            state_index = "attract"
            startRattle()

    elif state_index == "attract":
        rattle_duration += 1
        # if person is close,
        if isPersonClose():
            print("PERSON IS CLOSE!")
            # stop rattling to await open
            if isRattleDone():
                rattle_duration = 0
                print("STOPPING RATTLING TO AWAIT OPEN!")
                stopRattle()
                state_index = "wait_for_open"
        # if no motion and rattle is done, return to idle w/ no rattling
        if isMotionDetected() and isRattleDone():
            rattle_duration = 0
            print("PERSON STOPPED MOVING OR IS GONE!")
            stopRattle()
            state_index = "idle"

    elif state_index == "wait_for_open":
        wait_for_open_duration += 1
        # if lid is opened, move on to detecting if lid is closed
        if isLidOpened():
            wait_for_open_duration = 0
            print("LID WAS OPENED")
            state_index = "lid_open"
        # if no motion and rattle is done, return to idle w/ no rattling
        if isMotionDetected() and isWaitForOpenDone():
            wait_for_open_duration = 0
            print("PERSON STOPPED MOVING OR IS GONE!")
            state_index = "idle"

    elif state_index == "lid_open":
        # if lid is closed, begin scare routine! after duration, move on to stop
        if isLidClosed():
            fastOpenHinge()
            startRattle()
            time.sleep(SCARE_DURATION)
            state_index = "stop_scare"

    elif state_index == "stop_scare":
        # stop scare (rattle + close lid), then wait a duration
        stopRattle()
        slowCloseHinge()
        time.sleep(SLEEP_AFTER_SCARE)
        state_index = "wait_for_person_leave"

    elif state_index == "wait_for_person_leave":
        # after person is far, restart loop
        if isPersonClose():
            state_index = "idle"

    time.sleep(0.1)
